@@ -23,6 +23,7 @@ if subprocess.mswindows and sys.version_info[0:2] < (2,5):
 
 import pyutilib.services
 from pyutilib.common import *
+from pyutilib.misc import quote_split
 
 
 
@@ -151,13 +152,23 @@ def X_read_data(stream, ostr1, ostr2):
 # After this is finished, we can get the output from this command from
 # the process.stdout file descriptor.
 #
-def run_command(cmd, outfile=None, cwd=None, ostream=None, stdout=None, stderr=None, valgrind=False, valgrind_log=None, valgrind_options=None, memmon=False, env=None, define_signal_handlers=True, debug=False, verbose=True, timelimit=None, tee=None, ignore_output=False, shell=True, thread_reader=_read_data):
+def run_command(cmd, outfile=None, cwd=None, ostream=None, stdin=None, stdout=None, stderr=None, valgrind=False, valgrind_log=None, valgrind_options=None, memmon=False, env=None, define_signal_handlers=True, debug=False, verbose=True, timelimit=None, tee=None, ignore_output=False, shell=False, thread_reader=_read_data):
     #
     # Move to the specified working directory
     #
     if cwd is not None:
         oldpwd = os.getcwd()
         os.chdir(cwd)
+
+    cmd_type = type(cmd)
+    if cmd_type is list:
+        # make a private copy of the list
+        _cmd = cmd[:]
+    elif cmd_type is tuple:
+        _cmd = list(cmd)
+    else:
+        _cmd = quote_split(cmd.strip())
+        
     #
     # Setup memmoon
     #
@@ -165,7 +176,7 @@ def run_command(cmd, outfile=None, cwd=None, ostream=None, stdout=None, stderr=N
         memmon = pyutilib.services.registered_executable("memmon")
         if memmon is None:
             raise IOError, "Unable to find the 'memmon' executable"
-        cmd = memmon.get_path()+" "+cmd
+        _cmd.insert(0, memmon.get_path())
     #
     # Setup valgrind
     #
@@ -174,16 +185,20 @@ def run_command(cmd, outfile=None, cwd=None, ostream=None, stdout=None, stderr=N
         # The valgrind_log option specifies a logfile that is used to store
         # valgrind output.
         #
-        valgrind = pyutilib.services.registered_executable("valgrind")
-        if valgrind is None:
+        valgrind_cmd = pyutilib.services.registered_executable("valgrind")
+        if valgrind_cmd is None:
             raise IOError, "Unable to find the 'valgrind' executable"
+        valgrind_cmd = [ valgrind_cmd.get_path() ]
         if valgrind_options is None:
-            valgrind_options = " -v --tool=memcheck --trace-children=yes"
-        if not valgrind_log is None:
-            valgrind_log = " --log-file-exactly="+valgrind_log
+            valgrind_cmd.extend(
+                ( "-v","--tool=memcheck","--trace-children=yes" ))
+        elif type(valgrind_options) in (list, tuple):
+            valgrind_cmd.extend(valgrind_options)
         else:
-            valgrind_log = ""
-        cmd = valgrind.get_path()+valgrind_options+valgrind_log+" "+cmd
+            valgrind_cmd.extend(quote_split(valgrind_options.strip()))
+        if valgrind_log is not None:
+            valgrind_cmd.append("--log-file-exactly="+valgrind_log.strip())
+        _cmd = valgrind_cmd + _cmd
     #
     # Redirect stdout and stderr
     #
@@ -229,14 +244,14 @@ def run_command(cmd, outfile=None, cwd=None, ostream=None, stdout=None, stderr=N
             signal.signal(signal.SIGTERM, signal_handler)
     rc = -1
     if debug:
-        print "Executing command '%s'" % cmd
+        print "Executing command %s" % (_cmd, )
     try:
         GlobalData.signal_handler_busy=False
         if (tee is None) or (tee is False):
             #
             # Redirect IO to the stdout_arg/stderr_arg files
             #
-            process = SubprocessMngr(cmd.strip(), stdout=stdout_arg, stderr=stderr_arg, env=env, shell=shell)
+            process = SubprocessMngr(_cmd, stdin=stdin, stdout=stdout_arg, stderr=stderr_arg, env=env, shell=shell)
             GlobalData.current_process = process.process
             rc = process.wait(timelimit)
             GlobalData.current_process = None
@@ -249,7 +264,7 @@ def run_command(cmd, outfile=None, cwd=None, ostream=None, stdout=None, stderr=N
             rpipe1, wpipe1 = os.pipe()
             ##rpipe2, wpipe2 = os.pipe()
             #
-            process = SubprocessMngr(cmd.strip(), stdout=wpipe1, stderr=wpipe1, env=env, shell=shell)
+            process = SubprocessMngr(_cmd, stdin=stdin, stdout=wpipe1, stderr=wpipe1, env=env, shell=shell)
             GlobalData.current_process = process.process
             #
             # Create a thread to read in stdout and stderr data
@@ -282,7 +297,9 @@ def run_command(cmd, outfile=None, cwd=None, ostream=None, stdout=None, stderr=N
             ##os.close(rpipe2)
 
     except WindowsError, err:
-        raise ApplicationError, "Could not execute the command: "+cmd+"\tError message: "+str(err)
+        raise ApplicationError(
+            "Could not execute the command: '%s'\n\tError message: %s"
+            % (' '.join(_cmd), err) )
     except OSError:
         #
         # Ignore IOErrors, which are caused by interupts
@@ -319,7 +336,7 @@ else:
 
 class SubprocessMngr(object):
 
-    def __init__(self, cmd, stdout=None, stderr=None, env=None, bufsize=0, shell=False):
+    def __init__(self, cmd, stdin=None, stdout=None, stderr=None, env=None, bufsize=0, shell=False):
         """
         Setup and launch a subprocess
         """
@@ -329,6 +346,17 @@ class SubprocessMngr(object):
         #
         if stderr is None:
             stderr=subprocess.STDOUT
+
+        self.stdin = stdin
+        if stdin is None:
+            stdin_arg = None
+        else:
+            stdin_arg = subprocess.PIPE
+        #
+        # We would *really* like to deal with commands in execve form
+        #
+        if type(cmd) not in (list, tuple):
+            cmd = quote_split(cmd.strip())
         #
         # Launch subprocess using a subprocess.Popen object
         #
@@ -338,17 +366,17 @@ class SubprocessMngr(object):
             #
             startupinfo = subprocess.STARTUPINFO()
             #startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
-            self.process = subprocess.Popen(cmd.strip(), stdout=stdout, stderr=stderr, startupinfo=startupinfo, env=env, bufsize=bufsize, shell=shell)
+            self.process = subprocess.Popen(cmd, stdin=stdin_arg, stdout=stdout, stderr=stderr, startupinfo=startupinfo, env=env, bufsize=bufsize, shell=shell)
         elif False:   # subprocess.jython:
             #
             # Launch from Jython
             #
-            self.process = subprocess.Popen(cmd.strip(), stdout=stdout, stderr=stderr, env=env, bufsize=bufsize, shell=shell)
+            self.process = subprocess.Popen(cmd, stdin=stdin_arg, stdout=stdout, stderr=stderr, env=env, bufsize=bufsize, shell=shell)
         else:
             #
             # Launch on *nix
             #
-            self.process = subprocess.Popen(cmd.strip(), stdout=stdout, stderr=stderr, preexec_fn=os.setsid, env=env, bufsize=bufsize, shell=shell)
+            self.process = subprocess.Popen(cmd, stdin=stdin_arg, stdout=stdout, stderr=stderr, preexec_fn=os.setsid, env=env, bufsize=bufsize, shell=shell)
 
     def X__del__(self):
         """
@@ -381,7 +409,8 @@ class SubprocessMngr(object):
         timelimit has passed.
         """
         if timelimit is None:
-            return self.process.wait()
+            self.process.communicate(input=self.stdin)
+            return self.process.returncode
         else:
             #
             # Wait timelimit seconds and then force a termination
@@ -391,6 +420,12 @@ class SubprocessMngr(object):
             if timelimit <= 0:
                 raise ValueError, "'timeout' must be a positive number"
             endtime = timer()+timelimit
+
+            # This might be dangerous: we *could* deadlock if the input
+            # is large...
+            if self.stdin is not None:
+                self.process.stdin.write(self.stdin)
+            
             while timer() < endtime:
                 status = self.process.poll()
                 if status is not None:
@@ -472,7 +507,8 @@ if __name__ == "__main__": #pragma:nocover
         print ""
 
         print "X"
-        [rc,output] = run_command("python -c \"while True: print '.'\"")
+        [rc,output] = run_command("python -c \"while True: print '.'\"",
+                                  timelimit=2)
         print "Y"
         #[rc,output] = run_command("python -c \"while True: print '.'\"")
         [rc,output] = run_command("python -c \"while True: print '.'\"", verbose=False)
